# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class TestGandivaNullLiteralNode < Test::Unit::TestCase
  def setup
    omit("Gandiva is required") unless defined?(::Gandiva)
  end

  def test_invalid_type
    return_type = Arrow::NullDataType.new
    message =
      "[gandiva][null-literal-node][new] " +
      "failed to create: <#{return_type}>"
    assert_raise(Arrow::Error::Invalid.new(message)) do
      Gandiva::NullLiteralNode.new(return_type)
    end
  end

  def test_return_type
    return_type = Arrow::BooleanDataType.new
    literal_node = Gandiva::NullLiteralNode.new(return_type)
    assert_equal(return_type, literal_node.return_type)
  end
end
