// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

#include "arrow/python/config.h"
#include "arrow/python/config_internal.h"

#include "arrow/util/config.h"

namespace arrow {
namespace py {

namespace {

const BuildInfo kBuildInfo = {
    PYARROW_BUILD_TYPE,
};

}  // namespace

namespace internal {

bool IsOpenTelemetryEnabled() {
#ifdef ARROW_WITH_OPENTELEMETRY
  return true;
#else
  return false;
#endif
}

bool IsThreadingEnabled() {
#ifdef ARROW_ENABLE_THREADING
  return true;
#else
  return false;
#endif
}

}  // namespace internal

const BuildInfo& GetBuildInfo() { return kBuildInfo; }

}  // namespace py
}  // namespace arrow