% Enumeration class representing Date Units.

% Licensed to the Apache Software Foundation (ASF) under one or more
% contributor license agreements.  See the NOTICE file distributed with
% this work for additional information regarding copyright ownership.
% The ASF licenses this file to you under the Apache License, Version
% 2.0 (the "License"); you may not use this file except in compliance
% with the License.  You may obtain a copy of the License at
%
%   http://www.apache.org/licenses/LICENSE-2.0
%
% Unless required by applicable law or agreed to in writing, software
% distributed under the License is distributed on an "AS IS" BASIS,
% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
% implied.  See the License for the specific language governing
% permissions and limitations under the License.
classdef DateUnit < uint8

    enumeration
        Day          (0)
        Millisecond  (1)
    end

    methods (Hidden)

        function ticks = ticksPerSecond(obj)
            import arrow.type.DateUnit
            switch obj
                case DateUnit.Millisecond
                    ticks = 1e3;
                otherwise
                    error("arrow:dateunit:UnsupportedTicksPerSecond", ...
                        "The ticksPerSecond method can only be called on a DateUnit of type Millisecond.");
            end
        end

    end

end
