# Automatically generated. Don't modify manually.
#
# Red FlatBuffers version: 0.0.4
# Declared by:             //SparseTensor.fbs
# Rooting type:            org.apache.arrow.flatbuf.Message (//Message.fbs)

require "flatbuffers"

module ArrowFormat
  module Org
    module Apache
      module Arrow
        module Flatbuf
          class SparseMatrixCompressedAxis < ::FlatBuffers::Enum
            ROW = register("Row", 0)
            COLUMN = register("Column", 1)
          end
        end
      end
    end
  end
end
