# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

class TestDayTimeIntervalArray < Test::Unit::TestCase
  def test_value
    day_millisecond = Arrow::DayMillisecond.new(3, 100)

    builder = Arrow::DayTimeIntervalArrayBuilder.new
    builder.append_value(day_millisecond)
    array = builder.finish
    assert_equal(day_millisecond, array.get_value(0))
  end

  def test_values
    first_day_millisecond = Arrow::DayMillisecond.new(1, 100)
    second_day_millisecond = Arrow::DayMillisecond.new(3, 100)

    builder = Arrow::DayTimeIntervalArrayBuilder.new
    builder.append_value(first_day_millisecond)
    builder.append_value(second_day_millisecond)
    array = builder.finish
    assert_equal([first_day_millisecond, second_day_millisecond], array.values)
  end  
end
