# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pathlib
import sys
import types


def _is_path_like(path):
    return isinstance(path, str) or hasattr(path, '__fspath__')


def _ensure_path(path):
    if isinstance(path, pathlib.Path):
        return path
    else:
        return pathlib.Path(_stringify_path(path))


def _stringify_path(path):
    """
    Convert *path* to a string or unicode path if possible.
    """
    if isinstance(path, str):
        return path

    # checking whether path implements the filesystem protocol
    try:
        return path.__fspath__()
    except AttributeError:
        pass

    raise TypeError("not a path-like object")


def _import_pandas():
    # ARROW-13425: hide the in-tree PyArrow from pandas so it doesn't try
    # to use a possibly-broken development build.
    # ARROW-49905: The original workaround sets ``sys.modules['pyarrow'] = None``,
    # but pandas >= 3.0 has a different import mechanism resulting in
    # ``AttributeError: 'NoneType' object has no attribute 'Array'``
    # on the first DataFrame construction (in ``pandas._libs.lib.is_pyarrow_array``).
    # Instead, we install a minimal stub where ``isinstance(obj, pa.Array)``
    # check returns False for every real object.
    if "pyarrow" not in sys.modules:
        stub = types.ModuleType("pyarrow")
        stub.Array = type("Array", (), {})
        stub.ChunkedArray = type("ChunkedArray", (), {})
        # pandas.compat.pyarrow parses ``pa.__version__`` at import time.
        stub.__version__ = "0.0.0"
        sys.modules["pyarrow"] = stub
    import pandas as pd
    return pd


def _get_module(obj, *, default=None):
    """
    Try to find the name of the module *obj* is defined on.
    """
    try:
        return obj.__module__
    except AttributeError:
        # Might be a method/property descriptor as generated by Cython,
        # look up the enclosing class.
        try:
            return obj.__objclass__.__module__
        except AttributeError:
            return default
